import numpy as np


class Catmull():
    def __init__(self, points: np.ndarray, tension = 1.0):
        self.points = points
        self.tension = tension


    def point(self, progress: float):
        i = int(progress)
        local_prog = progress - i
        
        if i < 1:
            i = 1
        elif i > len(self.points) - 3:
            i = len(self.points) - 3
        
        p0, p1, p2, p3 = self.points[i - 1:i + 3]
        
        basis = np.array([1, local_prog, local_prog**2, local_prog**3])
        
        geometry = np.array([
            [0, 1, 0, 0],
            [-self.tension, 0, self.tension, 0],
            [2 * self.tension, self.tension - 3, 3 - 2 * self.tension, -self.tension],
            [-self.tension, 2 - self.tension, self.tension - 2, self.tension]
        ])
        
        return np.dot(basis, np.dot(geometry, np.array([p0, p1, p2, p3])))


import math
import numpy as np


class CubicBezier():
    def __init__(self, points : np.ndarray):
        self.points = points

    
    def point(self, prog : float):
        prog_cp = 1 - prog

        basis = np.empty(self.points.shape[0])

        for i in range(self.points.shape[0]):
            basis[i] = math.comb(self.points.shape[0] - 1, i) * prog_cp ** (self.points.shape[0] - 1 - i) * prog ** i
        
        return basis @ self.points


import math

from .consts import *


# Sine easing
def sine_in(progress : float):
    return 1 - math.cos((progress * math.pi) / 2)


def sine_out(progress : float):
    return math.sin((progress * math.pi) / 2)


def sine_in_out(progress : float):
    return -(math.cos(math.pi * progress) - 1) / 2

    
# Poly easing
def poly_in(progress : float, index : float = 2):
    return progress ** index


def poly_out(progress : float, index : float = 2):
    return 1 - (1 - progress) ** index


def poly_in_out(progress : float, index : float = 2):
    return 2 ** (index - 1) * progress ** index if progress < 0.5 else 1 - (-2 * progress + 2) ** index / 2
    

# Expo easing
def expo_in(progress : float):
    return 0 if progress == 0 else 2 ** (10 * progress - 10)


def expo_out(progress : float):
    return 1 if progress == 1 else 1 - 2 ** (-10 * progress)


def expo_in_out(progress : float):
    return 0 if progress == 0 else 1 if progress == 1 else 2 ** (20 * progress - 10) / 2 if progress < 0.5 else (2 - 2 ** (-20 * progress + 10)) / 2


# Circ easing
def circ_in(progress : float):
    return 1 - (1 - progress * progress) ** 0.5


def circ_out(progress : float):
    return (1 - (progress - 1) * (progress - 1)) ** 0.5


def circ_in_out(progress : float):
    return (1 - (1 - (2 * progress) * (2 * progress)) ** 0.5) / 2 if progress < 0.5 else ((1 - (-2 * progress + 2) * (-2 * progress + 2)) ** 0.5 + 1) / 2


# Back easing
def back_in(progress : float, bounce_const : float = 1.701540198866823876026881):
    return (bounce_const + 1) * progress * progress * progress - bounce_const * progress * progress


def back_out(progress : float, bounce_const : float = 1.701540198866823876026881):
    return (bounce_const + 1) * (progress - 1) ** 3 + bounce_const * (progress - 1) * (progress - 1) + 1


def back_in_out(progress : float, bounce_const : float = 2.592388901516299780093093):
    first = ((bounce_const + 1) * (2 * progress) ** 3 - bounce_const * 2 * progress * 2 * progress) / 2
    second = (1 + (bounce_const + 1) * (2 * progress - 2) ** 3 + bounce_const * (2 * progress - 2) ** 2) / 2 + 0.5

    return first if progress < 0.5 else second


# Elas easing
def elas_in(progress: float):
    if progress == 0:
        return 0
    if progress == 1:
        return 1
    return -math.exp2(10 * progress - 10) * math.sin((progress * 10 - 10.75) * elas_const_1)


def elas_out(progress: float):
    if progress == 0:
        return 0
    if progress == 1:
        return 1
    return math.exp2(-10 * progress) * math.sin((progress * 10 - 0.75) * elas_const_1) + 1


def elas_in_out(progress: float):
    if progress == 0:
        return 0
    if progress == 1:
        return 1
    if progress < 0.5:
        return -(math.exp2(20 * progress - 10) * math.sin((20 * progress - 11.125) * elas_const_2)) / 2
    return (math.exp2(-20 * progress + 10) * math.sin((20 * progress - 11.125) * elas_const_2)) / 2 + 1


class Easer():
    def __init__(self, start_value : float, end_value : float):
        self.start_value = start_value
        self.end_value = end_value


    def value(self, function : callable, progress : float, *args):
        return function(progress, *args) * (self.end_value - self.start_value) + self.start_value